import turtle as t
from itertools import combinations
import math

# The corners of an equilateral triangle centered around the origin
corners = [
    (math.cos(math.radians(-30)), math.sin(math.radians(-30))),
    (math.cos(math.radians(210)), math.sin(math.radians(210))),
    (0, 1)
]

# A scale factor so that the drawn simplices come out a nice size. This may
# seems like a very strange number, which it is. It has this precise value for
# historical reasons.
SCALE_FACTOR = .003 * math.sqrt(3./2)

# helper function for is_edge
def are_same(things):
    """
    Check if all the elements of the given iterable are equal
    """
    thlist = list(things)
    return all(thlist[0] == x for x in thlist[1:])

def is_edge((p1, p2), (q1, q2), ps):
    """
    Check if the given edge is a boundary of the given point set
    """
    if len(ps) <= 1:
        return True
    if p1 == q1:
        return are_same(a1 < p1 for (a1, a2) in ps)
    else:
        slope = (q2 - p2) / (q1 - p1)
        return are_same(a2 > p2 + slope * (a1 - p1) for (a1, a2) in ps)

class Drawer(object):
    """
    "Abstract" Drawer class. Subclasses must implement finalize(),
    initialize(), and _draw().

    Users of any of these classes must call initialize() before doing anything
    else with them and call finalize() when they're done with them.
    """
    def __init__(self, coords=(1,2,3)):
        """
        The `coords' argument is for drawing four- or higher-dimensional
        cones, and indicates which three coordinates should be projected onto
        """
        self.coords = coords

    def draw_outline(self):
        # hack:
        tmp = self.coords
        self.coords = (1,2,3)
        self.draw_simplicial([[0,1,0,0], [0,0,1,0], [0,0,0,1]], color="red")
        self.coords = tmp

    def draw_simplicial(self, points, color="black"):
        """
        Draw the given set of points. The points here are the ones given by
        cddlib, which means the coordinates only start in the second entry of
        the array (the first being a flag).
        """
        coords = []
        for p in points:
            v = p[1:]
            s = 1. * sum(v) * SCALE_FACTOR
            # project the point onto the standard 2-simplex
            v = tuple(x / s for x in v)
            # pick out three coordinates for drawing
            v = tuple(v[i-1] for i in self.coords)
            # barycentric coordinates, python-style
            coords.append(tuple(sum(u*q[i] for (u, q) in zip(v, corners)) for i in (0,1)))
        for p, q in combinations(coords, 2):
            c = coords[:]
            c.remove(p)
            c.remove(q)
            if not is_edge(p, q, c):
                continue
            self._draw(p, q, coords, color)

class SvgDrawer(Drawer):
    def initialize(self):
        print '<?xml version="1.0" encoding="UTF-8" standalone="no"?>'
        print '<svg xmlns:svg="http://www.w3.org/2000/svg" '\
              'xmlns="http://www.w3.org/2000/svg" version="1.0" '\
              'width="475" height="419" >'

    def finalize(self):
        print '</svg>'

    def _draw(self, p, q, coords, color):
        print ('<line x1="%d" y1="%d" x2="%d" y2="%d" '
               'style="stroke:%s;stroke-width:1"/>') % (p[0]+240,-p[1]+280,q[0]+240,-q[1]+280, color)

class TurtleDrawer(Drawer):
    def initialize(self):
        t.hideturtle()
        t.tracer(0)

    def finalize(self):
        t.done()

    def _draw(self, p, q, coords, color):
        t.pen(pencolor=color)
        t.penup()
        t.goto(*p)
        t.pendown()
        t.goto(*q)
        t.penup()
        a, b = zip(*coords)
        n = len(coords)
        t.clearstamps()
        t.goto(sum(a) / n, sum(b) / n)
        t.stamp()

class NothingDrawer(Drawer):
    def initialize(self):
        pass

    def finalize(self):
        pass

    def draw_simplicial(self, *args, **kwargs):
        pass

# This function isn't actually used anywhere in the program any more. Now you
# might argue that there's no reason to keep it, because version control. BUT:
# I don't want us to forget that we can also draw fans projectively instead of
# projecting onto the standard two-simplex, and this remains here as proof that
# it can be and has been done.
def draw_projective(points):
    t.hideturtle()
    t.pen(speed=10)
    for p, q in combinations(points[:], 2):
        t.penup()
        if p[3] == 0:
            p, q = q, p
        if p[3] == 0:
            continue
        t.goto(p[1]*20/p[3], p[2]*20/p[3])
        t.pendown()
        if q[3] != 0:
            t.goto(q[1]*20/q[3], q[2]*20/q[3])
        else:
            t.goto(t.xcor() + q[1]*1000, t.ycor()+q[2]*1000)
            t.penup()
