#!/usr/bin/env python

import os.path
import logging
import argparse

from parse_surface import parse_surface_file
from half_edge import BadEdgeException
import draw
import search

def show_cone(surface, drawer):
    """
    Show the cone corresponding to the given surface and print out its
    coordinates.
    """
    poly = surface.generate_inequalities()
    ineqs = poly.get_inequalities()
    ext = poly.get_generators()
    for row in ext[:]:
        print ("(" + ', '.join(["%6s"]*(len(row)-1)) + ")") % row[1:],
        eqs = [i + 1 for i, ineq in enumerate(ineqs) if
                sum(a * u for (a, u) in zip(ineq, row)) == 0]
        print "Equalities: %s" % ", ".join(map(str, eqs))
    print "Flippables: %s" % map(list, search.flippable_edges(poly))
    drawer.draw_simplicial(ext[:])

def show_fan(filename, auto, drawer, allow_noncommute):
    """
    Display the fan contained in the given file. If `auto' is True, find the
    cones automatically. Otherwise, enter interactive mode.
    """
    with open(filename) as f:
        surface = parse_surface_file(f)

    if auto:
        search.search(surface, lambda s, p, e: drawer.draw_simplicial(e[:]),
                      allow_noncommute)
        return

    show_cone(surface, drawer)

    while True:
        print
        print "?",
        inp = raw_input().strip()
        if inp == "q":
            exit(0)
        try:
            for i in map(int, inp.split(" ")):
                try:
                    surface.flip(i)
                except ValueError:
                    print "Ignoring invalid edge designation %d" % i
                except BadEdgeException:
                    print "Skipping %d because not flippable" % i
        except ValueError:
            print "Invalid input %s" % inp

        print "flipped edges: %s" % ", ".join(map(str, surface.flipped))
        show_cone(surface, drawer)


logging.basicConfig(level=logging.INFO)

def parse_axes(string):
    try:
        axes = tuple(int(n) for n in string.split(','))
        if len(axes) != 3:
            raise ValueError
        return axes
    except ValueError:
        msg = 'axes must be three comma-separated integers'
        raise argparse.ArgumentTypeError(msg)


parser = argparse.ArgumentParser(description='Compute a secondary fan.')
parser.add_argument('filename',
                    help='the file containing the surrface description')
parser.add_argument('-i', dest='auto', action='store_false',
                    help='use interactive mode')
parser.add_argument('-s', dest='draw', action='store_const', const='svg',
                    help='draw the fan to standard output as an svg')
parser.add_argument('-t', dest='draw', action='store_const', const='turtle',
                    help='draw the fan using turtle graphics')
parser.add_argument('-c', dest='noncomm', action='store_false',
                    help='forbid simultaneous non-commuting flips in non-'
                    'interactive mode')
parser.add_argument('-a', '--axes', dest='axes', action='store',
                    type=parse_axes,
                    help='a comma-separated list of three integers; these are '
                    'the axes which should be projected onto when drawing '
                    'higher-dimensional fans')

args = parser.parse_args()

if args.draw == 'turtle':
    drawerc = draw.TurtleDrawer
elif args.draw == 'svg':
    drawerc = draw.SvgDrawer
else:
    drawerc = draw.NothingDrawer
if args.axes:
    drawer = drawerc(args.axes)
else:
    drawer = drawerc()
drawer.initialize()
if not args.auto or not args.noncomm:
    drawer.draw_outline()
show_fan(args.filename, args.auto, drawer, args.noncomm)
drawer.finalize()
