from collections import namedtuple

try:
    import cdd
except ImportError:
    import sys
    sys.stderr.write("""\
    You need the package pycddlib to run this software!

    You can install with it "pip install pycddlib". If you haven't alredy, it
    is recommended that you first install virtualenv and virtualenvwrapper, and
    install pycddlib in a fresh virtual environment.  That is, once you have
    virtualenv and virtualenvwrapper, do this:

    $ mkvirtualenv secondaryfan
    $ workon secondaryfan
    $ pip install pycddlib\n""")

    exit(1)

class Halfedge(object):
    def __init__(self, tail, head, length, **kwargs):
        """
        Create half-edge from tail to head with the given length. An opposite
        edge is also automatically constructed.

        In fact, a half-edge without a succ is nonsense and shouldn't exist,
        but unfortunately, this generally can't be assigned until
        post-construction
        """
        self._head = head
        self._tail = tail
        self._length = length
        if "_opp" in kwargs:
            self._opp = kwargs["_opp"]
        else:
            self._opp = Halfedge(head, tail, length, _opp = self)
        self._succ = None

    def __repr__(self):
        return "%d->%s(%d)" % (self.tail, self.head, self.length)

    def set_succs(self, succ, oppsucc):
        if self._succ is not None or self.opp._succ is not None:
            raise ValueError("succs can (and must) be set only once!")
        self._succ = succ
        self.opp._succ = oppsucc

    @property
    def tail(self):
        return self._tail

    @property
    def head(self):
        return self._head

    @property
    def succ(self):
        return self._succ

    @property
    def opp(self):
        return self._opp

    @property
    def length(self):
        return self._length

    def scale(self, factor):
        self._length *= factor
        self.opp._length *= factor

    def is_flippable(self):
        """
        Check that this half-edge lies between two distinc triangles, and
        can therefore be flipped
        """
        return self is not self.succ and self is not self.succ.succ and \
               self is not self.succ.opp and self is not self.succ.succ.opp

    def flip_length(self):
        if not self.is_flippable():
            raise Exception()
        e2 = self.opp
        c = self.succ
        d = c.succ
        a = e2.succ
        b = a.succ
        return (a.length*c.length + b.length*d.length)/self.length

    def flip(self, reverse=False):
        # The second indicates whether this edge should switch in the
        # opposite (to the default) orientation. This exists because after
        # switching twice, the orientation is reversed

        #   /|\
        # c/ | \b
        # /  |e \
        # \  |  /
        # d\ | /a
        #   \|/
        # e is going from bottom to top
        # The new edge is going from left to right
        # The four other edges go counter-clockwise
        if not self.is_flippable():
            raise ValueError("Edge not flippable")
        e2 = self.opp
        c = self.succ
        d = c.succ
        a = e2.succ
        b = a.succ
        if reverse:
            a, c = c, a
            b, d = d, b
        # TRANSFORM INTO f
        newlen = self.flip_length()
        self._length = newlen
        self.opp._length = newlen
        self._tail = c.head
        self._head = a.head
        self.opp._tail = a.head
        self.opp._head = c.head
        # set new succs
        self._succ = b
        self.opp._succ = d
        a._succ = self.opp
        b._succ = c
        c._succ = self
        d._succ = a
        return self

class BadEdgeException(Exception):
    def __init__(self, i, *args, **kwargs):
       super(BadEdgeException, self).__init__("Non-flippable edge %d" % i, *args, **kwargs)

class Surface(object):
    def __init__(self, n, edges):
        self.n = n
        self.edges = edges
        self._flipped = []

    @property
    def flipped(self):
        return tuple(self._flipped)

    def generate_inequalities(self):
        """
        Generate the inequality polytope for this surface

            k
           /|\
          / | \
        l/  |  \j
         \  |  /
          \ | /
           \|/
            i
        We loop over ik, going upward
        list holding rows of the H-matrix
        """
        ineqs = []
        for he in self.edges:
            ik = he.length
            kl = he.succ.length
            il = he.succ.succ.length
            ij = he.opp.succ.length
            jk = he.opp.succ.succ.length

            i = he.tail
            j = he.opp.succ.head
            k = he.head
            l = he.succ.head

            ul = ik/(kl*il)
            uj = ik/(ij*jk)
            ui = - (kl/(il*ik) + jk/(ij*ik))
            uk = - (il/(kl*ik) + ij/(jk*ik))

            ineq = [0] * (self.n + 1)
            ineq[i] -= ui
            ineq[j] -= uj
            ineq[k] -= uk
            ineq[l] -= ul
            ineqs.append(ineq)
        mat = cdd.Matrix(ineqs, number_type='fraction')
        mat.rep_type = cdd.RepType.INEQUALITY
        return cdd.Polyhedron(mat)

    def flip(self, i):
        """
        Flip the ith edge (indexed from 1)
        """
        if int(i) != i or i < 1 or i > len(self.edges):
            raise ValueError("Invalid argument %d; edge must be an int between 1 and %d"
                    % (i, len(self.edges)))
        if self._flipped and self._flipped[-1] == i:
            self.edges[i-1].flip(True)
            del self._flipped[-1]
        else:
            try:
                self.edges[i-1].flip()
            except ValueError:
                raise BadEdgeException(i)
            self._flipped.append(i)

    def commute(self, i, j):
        """
        Test if edges i and j commute
        """
        e1 = self.edges[i-1]
        e2 = self.edges[j-1]
        neighbours = [e1.succ,
                      e1.succ.opp,
                      e1.succ.succ,
                      e1.succ.succ.opp,
                      e1.opp.succ,
                      e1.opp.succ.opp,
                      e1.opp.succ.succ,
                      e1.opp.succ.succ.opp]
        return all(e2 is not x for x in neighbours)

    def normalize(self):
        """
        Modify the lengths of the edges so that each c_ij^k=1.
        """
        # We loop over ik, as above, and add the angles c_kl^i and c_ij^k to
        # the running totals.
        horolengths = [0] * (self.n + 1)
        for he in self.edges:
            ik = he.length
            kl = he.succ.length
            il = he.succ.succ.length
            ij = he.opp.succ.length
            jk = he.opp.succ.succ.length

            horolengths[he.tail] += kl / (ik * il)
            horolengths[he.head] += ij / (ik * jk)

        invhl = [0] + [1/x for x in horolengths[1:]]

        for he in self.edges:
            he.scale(invhl[he.head] * invhl[he.tail])
