"""
This module parses the description of a surface and creates a half-edge complex
out of it.

The input file should consist of lines of the form

N a b succ oppsucc len

each representing a half-edge, where the first column consists of sequentially
increasing numbers (starting with 1), the second and third columns indicates
the start and end vertices of the half-edge, the fourth column indicates the
successor of the given half-egde, and the fifth the sucsessor of its opposite.
These are each indicated by the index given in the first column, whereby a
hypen ("-") should be prepended to indicate the opposite edge (e.g. "-4" means
the opposite of the edge on the fourth line). The last is a rational (i.e.
anything which can be parsed by the fractions.Fraction constructor) indicating
the length.

Note that each *pair* of half-edges is represented by one line, so there should
be as many lines as edges in the triangulation.

Blank lines and lines starting with a hash are ignored.
"""
from fractions import Fraction
from half_edge import Halfedge, Surface

class ParseError(Exception):
    pass

# The types of the fields are defined by this list of constructors
fields = [int] * 5 + [Fraction]

# Takes a file-like object corresponding to the above specification, and parses
# the living bejeesus out of it.
# If something goes wrong with parsing, a ParseError is raised, describing the
# problem.
# If the data is inconsistent, a ConsistencyError is raised, describing the
# problem
# The number of vertices and a list with the half-edges is returned
def parse_surface_file(file_object):
    # initialize the data
    hes = []
    succs = []
    # count the number of vertices
    num_vert = 0
    for line in file_object:
        line = line.strip()
        if line == "" or line.startswith("#"):
            continue
        try:
            data = [x for x in line.split(' ') if x]
            n, tail, head, succ, oppsucc, l = \
                    [cons(val) for cons, val in zip(fields, data)]
            if n != len(hes) + 1:
                raise ParseError("Incorrectly numbered line %d" % n)
            num_vert = max(num_vert, tail, head)
            he = Halfedge(tail, head, l)
            hes.append(he)
            succs.append((succ, oppsucc))
        except ValueError:
            raise ParseError("Problem parsing line `%s'" % line)
    for he, (succ, oppsucc) in zip(hes, succs):
        he.set_succs(hes[succ-1]    if succ    > 0 else hes[-succ-1].opp,
                     hes[oppsucc-1] if oppsucc > 0 else hes[-oppsucc-1].opp)

    surface = Surface(num_vert, hes)
    surface.normalize()
    return surface
