from half_edge import BadEdgeException

from itertools import combinations
import random
from fractions import Fraction
import logging

LOG = logging.getLogger("search")

def flippable_edges(poly):
    """
    Find the facets of the given polytope. We "assume" here that each
    half-plane intersects precisely the vertices of one facet (assuming that it
    intersects any at all). Therefore, we return, for each inequality, all the
    vertices at which that inequality is satisfied.

    Then, for each facet, return all inequalities which are simultaneously
    active at all of its vertices.
    """
    ineqs = poly.get_inequalities()
    vertices = poly.get_generators()
    dim = len(vertices[0]) - 1
    active_inqs_at_vs = [[] for i in range(len(vertices))]
    facets = set()
    for i, ineq in enumerate(ineqs):
        face = []
        for j, v in enumerate(vertices[:]):
            if sum(a * u for (a, u) in zip(ineq, v)) == 0:
                face.append(j)
                active_inqs_at_vs[j].append(i)
        if len(face) >= dim - 1:
            facets.add(frozenset(face))
    return [set.intersection(*[set(n + 1 for n in active_inqs_at_vs[i]) for i in f])
            for f in facets]

def normalize_points(points):
    normed = []
    for p in points:
        s = Fraction(sum(p))
        normed.append(tuple(x / s for x in p))
    return tuple(sorted(normed))

def d(msg, depth):
    return "-"*depth + msg

def search(surface, cb=None, allow_noncommute=True, found=set(), depth=0):
    poly = surface.generate_inequalities()
    ext = poly.get_generators()
    if len(ext) == 1:
        LOG.info(d("One-point degeneracy. Returning.", depth))
        return
    if len(ext) == 2:
        LOG.info(d("Two-point degeneracy. Returning.", depth))
        return
    ps = normalize_points(ext[:])
    if ps in found:
        LOG.info(d("Polytope already found. Returning.", depth))
        return
    if cb:
        # give callback redundant arguments for convenience
        cb(surface, poly, ext)
    ess = flippable_edges(poly)
    for es in ess:
        es = list(es)
        if sorted(surface.flipped[-len(es):]) == sorted(es):
            continue
        commute = True
        for a, b in combinations(es, 2):
            if not surface.commute(a, b):
                LOG.warning(d("Non-commuting edges %d and %d" % (a, b), depth))
                commute = False
        if not commute and not allow_noncommute:
            LOG.info("Skipping non-commuting flips")
            continue
        try:
            for i, e in enumerate(es):
                LOG.info(d("Flipping edge %d" % e, depth))
                surface.flip(e)
        except BadEdgeException:
            LOG.info(d("Non-flippable edge %d. Skipping." % e, depth))
            for e in reversed(es[:i]):
                surface.flip(e)
            continue
        found.add(ps)
        search(surface, cb, allow_noncommute, found, depth+1)
        for e in reversed(es):
            surface.flip(e)
